import typing
from abc import ABC, abstractmethod


class AbstractCaster(ABC):
    """
    Abstract caster class.
    """

    @abstractmethod
    def cast(self, value: str) -> typing.Any:
        """
        Cast the provided value to the desired type.
        :param value: value to be casted.
        :return: the casted value.
        """

        pass


class IntCaster(AbstractCaster):
    """
    Caster that casts the provided value to an integer.
    """

    def cast(self, value: str) -> int:
        return int(value)


class FloatCaster(AbstractCaster):
    """
    Caster that casts the provided value to a float.
    """

    def cast(self, value: str) -> float:
        return float(value)


class BoolCaster(AbstractCaster):
    """
    Caster that casts the provided value to a boolean.
    """

    true = {"true", "yes", "on", "enable", "1"}
    false = {"false", "no", "off", "disable", "0"}

    def __init__(self):
        pass

    def cast(self, value: str) -> bool:
        if value in self.true:
            return True
        elif value in self.false:
            return False

        raise ValueError("the provided value cannot be casted to bool")


class ListCaster(AbstractCaster):
    """
    Caster that casts the provided value to a list.
    """

    def __init__(self, sep: str = ",", item_caster: AbstractCaster | None = None):
        """
        Initialize the list caster.
        :param sep: separator used to split the provided value.
        :param item_caster: caster used to cast list items.
        """

        self.sep = sep
        self.item_caster = item_caster

    def cast(self, value: str) -> list[typing.Any]:
        str_items = value.split(self.sep)

        if self.item_caster:
            casted_items: list[typing.Any] = []
            for item in str_items:
                try:
                    casted = self.item_caster.cast(item)
                    casted_items.append(casted)
                except Exception as e:
                    raise ValueError(
                        f'failed to cast list item "{item}" using {self.item_caster.__class__.__name__}'
                    ) from e
            return casted_items

        return str_items



from .caster import AbstractCaster
from .provider import AbstractProvider


class CastingError(Exception):
    """
    Exception raised when an error occurs during casting.
    """

    field_name: str
    raw_value: str
    caster: AbstractCaster
    exception: Exception

    def __init__(
        self,
        field_name: str,
        raw_value: str,
        caster: AbstractCaster,
        exception: Exception,
        file_field_name: str | None = None,
        file_field_value: str | None = None,
    ):
        self.field_name = field_name
        self.raw_value = raw_value
        self.caster = caster
        self.exception = exception

        caster_name = caster.__class__.__name__

        if file_field_name:
            msg = (
                f'exception occurred when casting {field_name} raw value "{raw_value}" '
                f"using {caster_name}: {exception} (the raw value was read from {file_field_value}, which was provided by the attached file field {file_field_name})"
            )
        else:
            msg = (
                f'exception occurred when casting {field_name} raw value "{raw_value}" '
                f"using {caster_name}: {exception}"
            )
        super().__init__(msg)


class FieldValueNotProvidedError(Exception):
    """
    Exception raised when a field value is not provided by the provider.
    """

    field_name: str
    provider: AbstractProvider
    file_field_name: str | None

    def __init__(self, field_name: str, provider: AbstractProvider, file_field_name: str | None = None):
        self.field_name = field_name
        self.provider = provider
        self.file_field_name = file_field_name

        provider_name = provider.__class__.__name__
        if file_field_name is not None:
            msg = f"neither {field_name} nor {file_field_name} were provided by {provider_name}, at least one was expected"
        else:
            msg = f"{field_name} was not provided by {provider_name}, but was expected"

        super().__init__(msg)


class FieldConflictError(Exception):
    """
    Exception raised when a conflict occurs between a field and its corresponding file field.
    """

    field_name: str
    file_field_name: str

    def __init__(self, field_name: str, file_field_name: str, provider: AbstractProvider):
        self.field_name = field_name
        self.file_field_name = file_field_name

        provider_name = provider.__class__.__name__
        super().__init__(
            f"values for both {field_name} and {file_field_name} "
            f"fields were provided by {provider_name}, "
            "expected only value for one of them"
        )



import unittest
from unittest.mock import Mock, patch

from minicfg.field import Field, FieldConflictError, CastingError, FieldValueNotProvidedError, _NOT_SET
from minicfg.provider import AbstractProvider
from minicfg.caster import AbstractCaster


class TestField(unittest.TestCase):

    def test_field_initialization(self):
        field = Field(name="test_field", default="default_value", attach_file_field=True)
        self.assertEqual(field.name, "test_field")
        self.assertEqual(field._default, "default_value")
        self.assertTrue(field._attach_file_field)
        self.assertEqual(field._populated_value, _NOT_SET)

    def test_field_name_property(self):
        field = Field(name="test_field")
        self.assertEqual(field.name, "test_field")
        field.name = "new_name"
        self.assertEqual(field.name, "new_name")

    def test_field_populated_value_property(self):
        field = Field()
        self.assertEqual(field.populated_value, _NOT_SET)
        field._populated_value = "populated_value"
        self.assertEqual(field.populated_value, "populated_value")

    def test_populate_with_raw_value(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = "raw_value"
        field = Field(name="test_field")
        field.populate(provider)
        self.assertEqual(field.populated_value, "raw_value")

    def test_populate_with_file_field(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.side_effect = lambda key: "file_path" if key == "test_field_FILE" else None
        field = Field(name="test_field", attach_file_field=True)
        with patch("minicfg.field._read_raw_value_from_file", return_value="file_value"):
            field.populate(provider)
        self.assertEqual(field.populated_value, "file_value")

    def test_populate_with_default_value(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = None
        field = Field(name="test_field", default="default_value")
        field.populate(provider)
        self.assertEqual(field.populated_value, "default_value")

    def test_populate_raises_field_conflict_error(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.side_effect = lambda key: "raw_value" if key == "test_field" else "file_value"
        field = Field(name="test_field", attach_file_field=True)
        with self.assertRaises(FieldConflictError):
            field.populate(provider)

    def test_populate_raises_casting_error(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = "raw_value"
        caster = Mock(spec=AbstractCaster)
        caster.cast.side_effect = Exception("casting error")
        field = Field(name="test_field", caster=caster)
        with self.assertRaises(CastingError):
            field.populate(provider)

    def test_populate_raises_field_value_not_provided_error(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = None
        field = Field(name="test_field")
        with self.assertRaises(FieldValueNotProvidedError):
            field.populate(provider)


if __name__ == "__main__":
    unittest.main()


import unittest

from minicfg.caster import BoolCaster, FloatCaster, IntCaster, ListCaster


class TestIntCaster(unittest.TestCase):
    def setUp(self):
        self.caster = IntCaster()

    def test_positive(self):
        result = self.caster.cast("123123")
        self.assertEqual(result, 123123)

    def test_negative(self):
        result = self.caster.cast("-123123")
        self.assertEqual(result, -123123)

    def test_non_numeric(self):
        with self.assertRaises(ValueError):
            self.caster.cast("abcdef")

    def test_empty(self):
        with self.assertRaises(ValueError):
            self.caster.cast("")


class TestFloatCaster(unittest.TestCase):
    def setUp(self):
        self.caster = FloatCaster()

    def test_positive(self):
        result = self.caster.cast("123.456")
        self.assertEqual(result, 123.456)

    def test_negative(self):
        result = self.caster.cast("-123.456")
        self.assertEqual(result, -123.456)

    def test_non_numeric(self):
        with self.assertRaises(ValueError):
            self.caster.cast("abcdef")

    def test_empty(self):
        with self.assertRaises(ValueError):
            self.caster.cast("")


class TestBoolCaster(unittest.TestCase):
    def setUp(self):
        self.caster = BoolCaster()

    def test_true_values(self):
        true_values = ["true", "yes", "on", "enable", "1"]
        for value in true_values:
            self.assertTrue(self.caster.cast(value))

    def test_false_values(self):
        false_values = ["false", "no", "off", "disable", "0"]
        for value in false_values:
            self.assertFalse(self.caster.cast(value))

    def test_invalid_value(self):
        with self.assertRaises(ValueError):
            self.caster.cast("invalid")


class TestListCaster(unittest.TestCase):
    def setUp(self):
        self.separator = ","
        self.caster = ListCaster(sep=self.separator)

    def test_list_of_strings(self):
        result = self.caster.cast("a,b,c")
        self.assertEqual(result, ["a", "b", "c"])

    def test_list_with_int_caster(self):
        int_caster = IntCaster()
        list_caster = ListCaster(sep=self.separator, item_caster=int_caster)
        result = list_caster.cast("1,2,3")
        self.assertEqual(result, [1, 2, 3])

    def test_list_with_invalid_item(self):
        int_caster = IntCaster()
        list_caster = ListCaster(sep=self.separator, item_caster=int_caster)
        with self.assertRaises(ValueError):
            list_caster.cast("1,invalid,3")


if __name__ == "__main__":
    unittest.main()



import typing

from .field import Field
from .provider import AbstractProvider, EnvProvider

_DEFAULT_PROVIDER = EnvProvider


class Minicfg:
    """
    Base class for configuration classes.
    """

    _prefix: str  # prefix used for all fields in the Minicfg instance

    def __init__(self):
        if not hasattr(self, "_prefix"):
            # if prefix is not set, set it to an empty string:
            self._prefix = ""

    @classmethod
    def populated(cls, provider: AbstractProvider | None = None) -> "Minicfg":
        """
        Create an instance of the Minicfg class and populate it with the given provider.

        :param provider: provider used to populate the Minicfg instance.
        :return: populated Minicfg instance.
        """

        minicfg = cls()
        minicfg.populate(provider)
        return minicfg

    @property
    def prefix(self) -> str:
        return self._prefix

    @prefix.setter
    def prefix(self, value: str):
        self._prefix = value

    def populate(self, provider: AbstractProvider | None = None) -> None:
        """
        Populate the Minicfg instance using the given provider.
        All fields and child Minicfg instances will be populated recursively.

        :param provider: provider used to populate the Minicfg instance. If not provided, the default _DEFAULT_PROVIDER will be used.
        """

        if not provider:
            provider = _DEFAULT_PROVIDER()

        for attr_name, attr in self._iter_public_attrs():
            if isinstance(attr, type) and issubclass(attr, Minicfg):  # if attribute is a child Minicfg type
                # create an instance of the child minicfg:
                child_minicfg = attr()

                # use class name as a child minicfg prefix if it is not set:
                if child_minicfg.prefix == "":
                    child_minicfg.prefix = f"{child_minicfg.__class__.__name__}_"

                # prepend prefix from the parent minicfg instance:
                child_minicfg.prefix = f"{self._prefix}{child_minicfg._prefix}"

                # populate the child minicfg:
                child_minicfg.populate(provider)

                # replace the child minicfg class with its instance:
                self.__setattr__(attr_name, child_minicfg)

            elif isinstance(attr, Field):  # if attribute is an instance of a Field
                field = attr

                # if field name is not set, set it to the attribute name:
                if not field.name:
                    field.name = attr_name

                # populate field:
                field.populate(provider, self._prefix)

                # replace the field with its populated value:
                self.__setattr__(attr_name, field.populated_value)

    def _iter_public_attrs(self) -> typing.Generator[tuple[str, typing.Any], None, None]:
        """
        Iterate over public attributes of the Minicfg instance.

        :return: generator of tuples (attribute name, attribute).
        """

        for attr_name in dir(self):
            if attr_name.startswith("_"):
                continue

            attr = super().__getattribute__(attr_name)
            yield attr_name, attr


def minicfg_prefix(prefix: str):
    """
    Decorator for setting a prefix for the Minicfg class.
    Will be used as a prefix for all fields in the Minicfg class.

    Please note, that an "_" will be appended to the prefix. Use raw_prefix instead if you don't want it.
    :param prefix: prefix.
    """

    def decorator(cls: Minicfg):
        cls._prefix = f"{prefix}_"
        return cls

    return decorator


def minicfg_raw_prefix(prefix: str):
    """
    Decorator for setting a prefix for the Minicfg class.
    Will be used as a prefix for all fields in the Minicfg class.
    :param prefix: prefix.
    """

    def decorator(cls: Minicfg):
        cls._prefix = prefix
        return cls

    return decorator



import os
from abc import ABC, abstractmethod


class AbstractProvider(ABC):
    """
    Abstract provider class.
    """

    @abstractmethod
    def get(self, key: str) -> str | None:
        """
        Get the value for the given key.
        :param key: key to get the value for.
        :return: value for the given key or None if the key is not found. Please note, that the return value is always a string.
        """
        pass


class EnvProvider(AbstractProvider):
    """
    A provider that reads values from environment variables.
    """

    def get(self, key: str) -> str | None:
        return os.getenv(key)



import unittest
from unittest.mock import Mock
from minicfg.field import Field, _NOT_SET
from minicfg.exceptions import CastingError, FieldConflictError, FieldValueNotProvidedError
from minicfg.provider import AbstractProvider
from minicfg.caster import AbstractCaster


class TestField(unittest.TestCase):
    def test_field_initialization_with_default_values(self):
        field = Field(name="test_field")
        self.assertEqual(field.name, "test_field")
        self.assertEqual(field._default, _NOT_SET)
        self.assertIsNone(field._caster)
        self.assertFalse(field._attach_file_field)
        self.assertEqual(field._populated_value, _NOT_SET)

    def test_field_populate_with_provider_value(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = "123"
        caster = Mock(spec=AbstractCaster)
        caster.cast.return_value = 123

        field = Field(name="test_field", caster=caster)
        field.populate(provider)

        self.assertEqual(field.populated_value, 123)
        provider.get.assert_called_with("test_field")
        caster.cast.assert_called_with("123")

    def test_field_populate_with_default_value(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = None

        field = Field(name="test_field", default=456)
        field.populate(provider)

        self.assertEqual(field.populated_value, 456)

    def test_field_populate_with_file_value(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.side_effect = lambda x: "file_path" if x == "test_field_FILE" else None

        field = Field(name="test_field", attach_file_field=True)
        with unittest.mock.patch("minicfg.field._read_raw_value_from_file", return_value="789"):
            field.populate(provider)

        self.assertEqual(field.populated_value, "789")

    def test_field_populate_with_file_value_raises_casting_error(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.side_effect = lambda x: "file_path" if x == "test_field_FILE" else None
        caster = Mock(spec=AbstractCaster)
        caster.cast.side_effect = ValueError("Invalid cast")

        field = Field(name="test_field", attach_file_field=True, caster=caster)
        with unittest.mock.patch("minicfg.field._read_raw_value_from_file", return_value="invalid_value"):
            with self.assertRaises(CastingError):
                field.populate(provider)

    def test_field_populate_raises_field_conflict_error(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.side_effect = lambda x: "value" if x == "test_field" else "file_path"

        field = Field(name="test_field", attach_file_field=True)
        with self.assertRaises(FieldConflictError):
            field.populate(provider)

    def test_field_populate_raises_casting_error(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = "invalid_value"
        caster = Mock(spec=AbstractCaster)
        caster.cast.side_effect = ValueError("Invalid cast")

        field = Field(name="test_field", caster=caster)
        with self.assertRaises(CastingError):
            field.populate(provider)

    def test_field_populate_raises_field_value_not_provided_error(self):
        provider = Mock(spec=AbstractProvider)
        provider.get.return_value = None

        field = Field(name="test_field")
        with self.assertRaises(FieldValueNotProvidedError):
            field.populate(provider)


if __name__ == "__main__":
    unittest.main()


from typing import Any

from .caster import AbstractCaster
from .exceptions import CastingError, FieldConflictError, FieldValueNotProvidedError
from .provider import AbstractProvider

_NOT_SET = object()


def _read_raw_value_from_file(path: str) -> str:
    """
    Read the raw value from the file at the given path.
    :param path: path to the file.
    :return: the raw value read from the file (with leading and trailing whitespaces removed).
    """

    with open(path, "r") as file:
        return file.read().strip()


class Field:
    """
    Field class represents a configuration field.
    """

    _name: str  # name of the field

    _default: Any  # default value of the field
    _caster: AbstractCaster  # caster used to cast raw values
    _attach_file_field: bool  # indicates whether file field should be attached to the field

    _populated_value: Any  # value determined after field population

    def __init__(
        self,
        name: str | None = None,
        default: Any = _NOT_SET,
        caster: AbstractCaster | None = None,
        attach_file_field: bool = False,
    ):
        self._name = name

        self._default = default
        self._caster = caster
        self._attach_file_field = attach_file_field

        self._populated_value = _NOT_SET

    @property
    def name(self) -> str | None:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def populated_value(self) -> Any:
        return self._populated_value

    def populate(self, provider: AbstractProvider, field_name_prefix: str | None = None) -> None:
        """
        Populate the field using the given provider.
        :param provider: provider used to get the field raw value.
        :param field_name_prefix: prefix to prepend to the field name.
        """

        if field_name_prefix is None:
            field_name_prefix = ""

        field_name = f"{field_name_prefix}{self._name}"  # name of the field

        raw_value_from_provider: str | None = provider.get(field_name)
        if raw_value_from_provider is not None:
            if self._attach_file_field:
                file_field_name = f"{field_name_prefix}{self._name}_FILE"  # name of the corresponding file field
                if provider.get(file_field_name) is not None:
                    raise FieldConflictError(field_name, file_field_name, provider)

            try:
                self._populated_value = self._cast_raw_value_if_needed(raw_value_from_provider)
                return
            except Exception as e:
                raise CastingError(
                    field_name=field_name, raw_value=raw_value_from_provider, caster=self._caster, exception=e
                ) from e

        if self._attach_file_field:
            file_field_name = f"{field_name_prefix}{self._name}_FILE"  # name of the corresponding file field
            filepath = provider.get(file_field_name)
            if filepath is not None:
                raw_value_from_file = _read_raw_value_from_file(filepath)
                try:
                    self._populated_value = self._cast_raw_value_if_needed(raw_value_from_file)
                    return
                except Exception as e:
                    raise CastingError(
                        field_name=field_name,
                        raw_value=raw_value_from_file,
                        caster=self._caster,
                        exception=e,
                        file_field_name=file_field_name,
                        file_field_value=filepath,
                    ) from e

            if self._default is not _NOT_SET:
                self._populated_value = self._default
                return

            raise FieldValueNotProvidedError(field_name, provider, file_field_name)

        if self._default is not _NOT_SET:
            self._populated_value = self._default
            return

        raise FieldValueNotProvidedError(field_name, provider)

    def _cast_raw_value_if_needed(self, raw_value: str) -> Any:
        """
        Cast the raw value using the caster if it is set.
        If no caster is set, return the raw value as is.
        :param raw_value: the raw value to be cast.
        :return: the casted value.
        """
        if self._caster:
            return self._caster.cast(raw_value)
        return raw_value
